#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include "version.h"

static void printUsage(void) {
	printf("Usage: agg [OPTION]... [FILE]\n"
	       "Aggregates group sums of the records from input\n"
	       "  -d DELIM  use DELIM instead of comma\n"
	       "  -g LIST   Define grouping fields. Default is first field\n"
	       "  -f LIST   Sum these fields. Default is second field\n"
	       "  -t        Trim (remove whitespace from fields)\n"
	       "  -h        Display help\n"
	      );
}


#define MAXFIELDS 128
struct Field {
	double dbl;
	int64_t i64;
	int type;
	char str[256];
};
#define FIELDTYPE_GROUP		0 /*grouping field*/
#define FIELDTYPE_STRING	1 /*normal field*/
#define FIELDTYPE_DOUBLE_SUM	2 /*field sum as double*/
#define FIELDTYPE_INT64_SUM	3 /*field sum as int64_t*/

/*Fields for the group (includes sum) */
static struct Field group_field[MAXFIELDS];
static int group_fields;
/*Fields in current record*/
static char current_field[MAXFIELDS][256];
static int current_fields;
/*Indexes to grouping fields*/
static int grouping_fields;
static int grouping_field_index[MAXFIELDS];
/*indexes to summing fields*/
static int summing_fields;
static int summing_field_index[MAXFIELDS];
/*max(summing_field_index[],grouping_field_index[]) */
static int max_significant_field_index;

static int lineno;
static char delimiter;
static int trim_fields;

static void initialize(void) {
	int i;
	for(i=0; i<MAXFIELDS; i++)
		group_field[i].type = FIELDTYPE_STRING;
	
	group_field[0].type = FIELDTYPE_GROUP;
	group_field[1].type = FIELDTYPE_INT64_SUM;
	lineno = 0;
	
	delimiter = ',';
	
	trim_fields = 0;
}


static void setup_field_indexes(void) {
	int i;
	grouping_fields = 0;
	summing_fields = 0;
	for(i=0; i<MAXFIELDS; i++) {
		if(group_field[i].type==FIELDTYPE_GROUP)
			grouping_field_index[grouping_fields++] = i;
		else if(group_field[i].type==FIELDTYPE_DOUBLE_SUM ||
		        group_field[i].type==FIELDTYPE_INT64_SUM
		       )
			summing_field_index[summing_fields++] = i;
	}
	if(summing_field_index[summing_fields-1] > grouping_field_index[grouping_fields-1])
		max_significant_field_index = summing_field_index[summing_fields-1];
	else
		max_significant_field_index = grouping_field_index[grouping_fields-1];
}


static int read_next_fields(FILE *fp) {
	static char line[4096];
	char *p = line;
	if(!fgets(line,sizeof(line),fp))
		return 0;
	lineno++;
	current_fields = 0;
	
	for(;;) {
		char *s,*e;
		size_t vlen;
		if(trim_fields)
			while(*p && isspace(*p)) p++;
		s = p;
		while(*p && *p!=delimiter)
			p++;
		e=p;
		if(trim_fields) {
			while(e>s+1 && isspace(e[-1]))
				e--;
		}
		
		vlen = e-s;
		if(*e=='\0' && e[-1]=='\n') vlen--;
		if(vlen>=256) {
			fprintf(stderr,"agg: line %d contained field longer than 256 characters\n",lineno);
			exit(EXIT_FAILURE);
		}
		memcpy(current_field[current_fields], s, vlen);
		current_field[current_fields][vlen] = '\0';
		current_fields++;
		
		if(*p) p++;
		else break;
	}
	if(current_fields<=max_significant_field_index) {
		fprintf(stderr,"agg: line %d contained less than %d fields\n", lineno, max_significant_field_index+1);
		exit(EXIT_FAILURE);
	}
	return 1;
}


static void finish_previous_group(void) {
	int i;
	if(lineno<=1) return;
	for(i=0; i<group_fields; i++) {
		switch(group_field[i].type) {
			case FIELDTYPE_GROUP:
			case FIELDTYPE_STRING:
				fputs(group_field[i].str,stdout);
				break;
			case FIELDTYPE_DOUBLE_SUM:
				printf("%f",group_field[i].dbl);
				break;
			case FIELDTYPE_INT64_SUM:
				printf("%"PRId64, group_field[i].i64);
				break;
		}
		if(i+1<group_fields)
			putchar(delimiter);
	}
	putchar('\n');
}


static void begin_new_group(void) {
	int i;
	for(i=0; i<current_fields; i++) {
		switch(group_field[i].type) {
			case FIELDTYPE_GROUP:
				break;
			case FIELDTYPE_STRING:
				break;
			case FIELDTYPE_DOUBLE_SUM:
				group_field[i].type = FIELDTYPE_INT64_SUM	;
			case FIELDTYPE_INT64_SUM:
				group_field[i].i64 = 0;
				break;
		}
		strcpy(group_field[i].str, current_field[i]);
	}
	group_fields = current_fields;
}


static void aggregate_fields(void) {
	int i;
	for(i=0; i<summing_fields; i++) {
		int j = summing_field_index[i];
		const char *value_str = current_field[j];
		int64_t i64;
		double dbl;
		char *endptr;
		
		endptr = 0;
		i64 = strtol(value_str,&endptr,0);
		if(endptr==0 || *endptr=='\0' || isspace(*endptr)) {
			if(group_field[j].type==FIELDTYPE_INT64_SUM)
				group_field[j].i64 += i64;
			else
				group_field[j].dbl += i64;
			continue;
		}
		endptr=0;
		dbl = strtod(value_str,&endptr);
		if(endptr==0 || *endptr=='\0' || isspace(*endptr)) {
			if(group_field[j].type==FIELDTYPE_INT64_SUM) {
				group_field[j].dbl = (double)group_field[j].i64; /*we may loose significant bits here*/
				group_field[j].type=FIELDTYPE_DOUBLE_SUM;
			}
			group_field[j].dbl += dbl;
			continue;
		}
		fprintf(stderr,"agg: line %d contained unsummable field '%s'\n",lineno,current_field[j]);
		exit(EXIT_FAILURE);
	}
}


static int is_new_group(void) {
	int i;
	if(lineno==1) return 1;
	for(i=0; i<grouping_fields; i++) {
		int j = grouping_field_index[i];
		if(strcmp(group_field[j].str, current_field[j])!=0)
			return 1;
	}
	return 0;
}


static void aggregate(FILE *fp) {
	for(;;) {
		if(!read_next_fields(fp)) {
			finish_previous_group();
			return;
		}
		if(is_new_group()) {
			finish_previous_group();
			begin_new_group();
		}
		aggregate_fields();
	}
}


static void parse_field_spec(const char *spec, int fieldtype) {
	int i;
	const char *p=spec;
	for(i=0; i<MAXFIELDS; i++)
		if(group_field[i].type==fieldtype)
			group_field[i].type=FIELDTYPE_STRING;
	while(*p) {
		const char *s;
		char *endptr;
		long from_fieldno,to_fieldno,f;
		
		while(isspace(*p)) p++;
		if(!*p) break;
		s=p;
		while(isdigit(*p)) p++;
		
		endptr=0;
		from_fieldno = strtol(s,&endptr,0);
		if(endptr!=p ||
		   from_fieldno<1 ||
		   from_fieldno>=MAXFIELDS)
		{
			fprintf(stderr,"agg: Invalid field specification: %s\n",spec);
			exit(EXIT_FAILURE);
		}
		
		
		if(*p=='-') {
			p++;
			s=p;
			while(isdigit(*p)) p++;
			endptr=0;
			to_fieldno = strtol(s,&endptr,0);
			if(endptr!=p ||
			   to_fieldno<1 ||
			   to_fieldno>=MAXFIELDS)
			{
				fprintf(stderr,"agg: Invalid field specification: %s\n",spec);
				exit(EXIT_FAILURE);
			}
		} else
			to_fieldno = from_fieldno;
		
		for(f=from_fieldno; f<=to_fieldno; f++)
			group_field[f-1].type = fieldtype;
		
		if(*p==',') p++;
	}
}


int main(int argc, char **argv) {
	int c;
	FILE *fp;
	
	initialize();
	
	while((c=getopt(argc,argv,"vd:g:f:th"))!=EOF) {
		switch(c) {
			case 'v':
				printf("agg "VERSION"\n");
				exit(EXIT_FAILURE);
			case 'd':
				delimiter = *optarg;
				break;
			case 'g':
				parse_field_spec(optarg,FIELDTYPE_GROUP);
				break;
			case 'f':
				parse_field_spec(optarg,FIELDTYPE_INT64_SUM);
				break;
			case 't':
				trim_fields = 1;
				break;
			case 'h':
				printUsage();
				exit(EXIT_FAILURE);
		}	
	}
	setup_field_indexes();
	if(optind<argc) {
		fp=fopen(argv[optind],"r");
		if(!fp) {
			perror(argv[optind]);
			exit(EXIT_FAILURE);
		}
	} else
		fp = stdin;
	setvbuf(fp,0,_IOFBF,16384);
	aggregate(fp);
	
	return EXIT_SUCCESS;
}
